import click
from data.main import connect
import os
from pathlib import Path
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

out_path = Path(os.getenv('DATA_MINING_DOC_DIR')) / 'figures'
@click.command('plot:sentiment-over-time')
def over_time():
    filename = "sentiment_over_time.png"

    DB = connect()
    data = DB.sql("""
        SELECT
            avg(sent.class_id) as sentiment
            ,s.published_at as date
        FROM top.story_sentiments sent
        JOIN top.stories s
        ON s.id = sent.story_id
        GROUP BY
            s.published_at
    """).df()
    DB.close()

    ax = sns.scatterplot(x=data['date'], y=data['sentiment'])
    ax.set(title="sentiment vs. time")
    plt.tight_layout()
    plt.savefig(out_path / filename)
    print(f"saved: {filename}")
@click.command('plot:bias-vs-sentiment-over-time')
def bias_over_time():
    filename = "bias_vs_sentiment_over_time.png"

    DB = connect()
    data = DB.sql("""
        SELECT
            avg(sent.class_id) as sentiment
            ,s.published_at as date
            ,b.id as bias_id
        FROM top.story_sentiments sent
        JOIN top.stories s
        ON s.id = sent.story_id
        JOIN publisher_bias pb
        ON pb.publisher_id = s.publisher_id
        JOIN bias_ratings b
        ON b.id = pb.bias_id
        GROUP BY
            s.published_at
            ,b.id
    """).df()
    DB.close()

    ax = sns.relplot(x=data['date'], y=data['sentiment'], col=data['bias_id'])
    ax.set(title="sentiment vs. time grouped by bias")
    plt.tight_layout()
    plt.savefig(out_path / filename)
    print(f"saved: {filename}")
